MobSpells = LibStub("AceAddon-3.0"):NewAddon("MobSpells")
local AceGUI = LibStub("AceGUI-3.0")

LibStub("AceAddon-3.0"):EmbedLibraries(
	MobSpells,
	"AceEvent-3.0",
	"AceConsole-3.0"
)

local defaults = {global = {mobs = {}}, profile = {showTooltips = true}}
local mobs
local bit_band = _G.bit.band
local frame

-- Upvalues
local next = _G.next
local select = _G.select
local setmetatable = _G.setmetatable
local pairs = _G.pairs
local type = _G.type
local table_concat = _G.table.concat
local table_sort = _G.table.sort
--[[
local COMBATLOG_OBJECT_CONTROL_NPC = _G.COMBATLOG_OBJECT_CONTROL_NPC
local COMBATLOG_OBJECT_TYPE_NPC = _G.COMBATLOG_OBJECT_TYPE_NPC
local COMBATLOG_OBJECT_REACTION_HOST = _G.COMBATLOG_OBJECT_REACTION_HOST
local COMBATLOG_OBJECT_REACTION_HOSTILE = _G.COMBATLOG_OBJECT_REACTION_HOSTILE
local COMBATLOG_DEFAULT_COLORS = _G.COMBATLOG_DEFAULT_COLORS
]]--
local tonumber = _G.tonumber
local UnitGUID = _G.UnitGUID
local GetZoneText = _G.GetZoneText
local GameTooltip = _G.GameTooltip
local GetSpellInfo = _G.GetSpellInfo
local IsInInstance = _G.IsInInstance
local GetInstanceDifficulty = _G.GetInstanceDifficulty
local GetTime = _G.GetTime
local IsShiftKeyDown = _G.IsShiftKeyDown
local ChatFrameEditBox = _G.ChatFrameEditBox
local GetSpellLink = _G.GetSpellLink
local UnitName = _G.UnitName
local SendChatMessage = _G.SendChatMessage
local GetNumRaidMembers = _G.GetNumRaidMembers
local tinsert, tremove = _G.tinsert, _G.tremove

-- Table recycling
local new, newHash, newSet, del
do
	local list = setmetatable({}, {__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	function newHash(...)
		local t = next(list)
		if t then
			list[t] = nil
		else
			t = {}
		end	
		for i = 1, select('#', ...), 2 do
			t[select(i, ...)] = select(i+1, ...)
		end
		return t
	end
	function newSet(...)
		local t = next(list)
		if t then
			list[t] = nil
		else
			t = {}
		end	
		for i = 1, select('#', ...) do
			t[select(i, ...)] = true
		end
		return t
	end
	function del(t)
		setmetatable(t, nil)
		for k in pairs(t) do
			t[k] = nil
		end
		t[''] = true
		t[''] = nil
		list[t] = true
		return nil
	end
end

function MobSpells:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("MobSpellsDB", defaults)

	mobs = self.db.global.mobs
	self:RegisterChatCommand("mobspells", "Show")
	self:RegisterChatCommand("spells", "GetTargetSpells")
	-- self:RegisterChatCommand("reportspells", "ReportTargetSpells")
	self.searchIndex = self.searchIndex or "NPC"
end

function MobSpells:OnEnable()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	if self.db.profile.showTooltips then
		self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
	end
end

--event which indicate the cast of a spell
--one cast of a spell should not produce more than one of these events
local start_events = {
    SWING_DAMAGE = true,
    SWING_MISSED = true,
    SPELL_CAST_START = true, 
    SPELL_CAST_SUCCESS = true, 
    SPELL_PERIODIC_CAST_START = true,
    SPELL_AURA_APPLIED = true,
}

local amount_suffixes = {
    ["ENERGIZE"] = true,
    ["LEECH"] = true,
    ["DRAIN"] = true,
}

local schools = {
    [1] = "Physical",
    [2] = "Holy",
    [4] = "Fire",
    [8] = "Nature",
    [16] = "Frost",
    [32] = "Shadow",
    [64] = "Arcane",
}

--[[
    Checks if source or target of the event is a mob we are interested in.
    Returns:
        the NPCID of the mob or nil
        the GUID of the mob
        boolean flag indicating if the mob is destination(true) or source(false) of the event
        the name of the mob
    ]]
function MobSpells:GetMobID(eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags)
    local guid, name, isDest
    
	if bit_band(srcFlags, COMBATLOG_OBJECT_CONTROL_NPC) == COMBATLOG_OBJECT_CONTROL_NPC and 
	   bit_band(srcFlags, COMBATLOG_OBJECT_TYPE_NPC) == COMBATLOG_OBJECT_TYPE_NPC then
       
        isDest = false
		guid = srcGUID
		name = srcName
        
	elseif bit_band(dstFlags, COMBATLOG_OBJECT_CONTROL_NPC) == COMBATLOG_OBJECT_CONTROL_NPC and
           bit_band(dstFlags, COMBATLOG_OBJECT_REACTION_HOSTILE) == COMBATLOG_OBJECT_REACTION_HOSTILE and 
           bit_band(dstFlags, COMBATLOG_OBJECT_TYPE_NPC) == COMBATLOG_OBJECT_TYPE_NPC then
        
        isDest = true
		guid = dstGUID
		name = dstName
	end
    
    if not guid then return end
    
	local NPCID = tonumber(guid:sub(-12,-7),16)		
    
    return NPCID, guid, isDest, name
end

do
	local tempTable = {}
	local DEFAULT_COLOR = {r = 1, g = 1, b = 1, a = 1}
	function MobSpells:UPDATE_MOUSEOVER_UNIT()
		local guid = UnitGUID("mouseover")
		local NPCID = tonumber(guid:sub(-12,-7),16)
		local zone = GetZoneText()
	    if IsInInstance() and GetInstanceDifficulty() == 2 then 
	        zone = zone.." [Heroic]" 
	    end
		
		local n, ln = 0, 0
		if mobs[zone] and mobs[zone][NPCID] and mobs[zone][NPCID].spells then
			for k, v in pairs(mobs[zone][NPCID].spells) do
				if type(v) == "number" then
					mobs[zone][NPCID].spells[k] = {school = v}
					v = mobs[zone][NPCID].spells[k]
				end
                if k ~= 0 then 
                    local name = GetSpellInfo(k)
                    
                    local color = COMBATLOG_DEFAULT_COLORS.schoolColoring[v.school] or COMBATLOG_DEFAULT_COLORS.schoolColoring.default
                    if not color then
                        color = DEFAULT_COLOR
                    end
                    local text = ("|c%02x%02x%02x%02x%s|r"):format(color.a * 255, color.r * 255, color.g * 255, color.b * 255, name)
                    n = n + 1
                    tempTable[n] = text
                    if n == 3 then
                        GameTooltip:AddLine((ln == 0 and "Casts " or "") .. table_concat(tempTable, ", ", 1, n))
                        GameTooltip:Show()
                        ln = ln + 1
                        n = 0
                    end
                end
			end
		end
		if n > 0 then
			GameTooltip:AddLine((ln == 0 and "Casts " or "") .. table_concat(tempTable, ", ", 1, n))
			GameTooltip:Show()
		end
	end
end
--suffixes of spells which affect other spells
local extra_spell_suffixes = {
    ["INTERRUPT"] = true,
    ["AURA_DISPELLED"] = true,
    ["AURA_STOLEN"] = true,
}

--[[
    Checks if the events contains information on an spell we are interested in.
    Returns: 
        the spell ID of the spell of interest or nil
        the start of the parameters of the spell-event
        the prefix of the event
        the suffix of the event
    ]]
function MobSpells:GetSpellID(eventtype, isDest, ...)
    local spellID, school
    
    local prefix = eventtype:sub(1, 5)
    
    if prefix == "ENVIR" then return end    
    if prefix == "SPELL" and eventtype:sub(7, 14) == "PERIODIC" then prefix = "SWING_PERIODIC" end
    
    local paramStart = 12
    if prefix == "SWING" then paramStart = 9 end
    
    local suffix = eventtype:sub(#prefix + 2)
    
    --if the mob is the destination of the spell we are only interested in it if the spell was a
    --spell which affected an aura of the mob
    if isDest and extra_spell_suffixes[suffix] then
        local spellIDTmp, _, schoolTmp, auraType = select(paramStart, ...) 
        if auraType and auraType == "BUFF" then 
            spellID = spellIDTmp 
            school = schoolTmp
        end
        
    --otherwise everything might be interresting
    elseif not isDest then
        if prefix == "SWING" then
            spellID = 0
        else
            spellID, _, school = select(9, ...) 
        end
    end
    
    return spellID, paramStart, prefix, suffix, school
end

function MobSpells:COMBAT_LOG_EVENT_UNFILTERED(event, ...)
	local timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags = ...
    
    --check if the events affects a mob we are interested in and return its guid (or nil)
    local NPCID, guid, isDest, name = self:GetMobID(eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags)	
    if not NPCID then return end
    
    --check if the events contains a spell we are interested in and return its id (or nil)
    local spellID, paramStart, prefix, suffix, school = self:GetSpellID(eventtype, isDest, ...)
    if not spellID then return end
    
    --get zone + difficulty
    local zone = GetZoneText()
    if IsInInstance() and GetInstanceDifficulty() == 2 then 
        zone = zone.." [Heroic]" 
    end
    
    mobs[zone] = mobs[zone] or {}
    mobs[zone][NPCID] = mobs[zone][NPCID] or {name = name, spells = {}}
    local spell = mobs[zone][NPCID].spells[spellID] or {}
	if type(spell) ~= "table" then
		local school = spell
		spell = {}
		spell.school = school
	end
    
    -----------start of parsing spell informations-----------
    
    --debug only: adds the eventtype to the spellinfo so we know which events affected a spell
    spell.events = spell.events or {}
    spell.events[eventtype] = spell.events[eventtype] or {}
    if type(spell.events[eventtype]) ~= "function" then spell.events[eventtype] = {} end
    
    spell.school = school
	
    --flags concerning the ability of the player to manipualte the spell
    if not spell.interruptable and suffix == "INTERRUPT" then spell.interruptable = true end
    if not spell.stealable and suffix == "AURA_STOLEN" then spell.stealable = true end
    if not spell.dispellable and suffix == "AURA_DISPELLED" then spell.dispellable = true end
    
    --information on time between casts of a spell
    if start_events[eventtype] then
        local time = GetTime()
        if spell.lastTime and spell.lastGUID == guid then
            local gone = time - spell.lastTime
            
            if not spell.minSpeed or spell.minSpeed > gone then
                spell.minSpeed = gone
            end
            if not spell.maxSpeed or spell.maxSpeed < gone then
                spell.maxSpeed = gone
            end
        end        
        spell.lastTime = time
        spell.lastGUID = guid
    end
    
    --informations concerning damage
    if suffix == "DAMAGE" then
        local amount, school, resisted, blocked, absorbed, critical, glancing, crushing = select(paramStart, ...)  
        spell.schoolName = schools[school]
        
        --flags on reduce-ability of the damage (resist and block)
        if resisted then 
            spell.resisted = true 
            amount = amount + resisted
        end
        if blocked then 
            spell.blocked = true 
            amount = amount + blocked
        end
        if absorbed then amount = amount + absorbed end    
        
        --flags on damage burst-ability
        if critical then spell.critical = true end
        if glancing then spell.glancing = true end
        if crushing then spell.crushing = true end
        
        --don't record bursted damage (for now)
        if not critical and not glancing and not crushing then
            local min = spell.amountMin
            if not min or min > amount then
               spell.amountMin = amount
            end
            
            local max = spell.amountMax
            if not max or max < amount then
               spell.amountMax = amount
            end
        end
        
    --information concerning heals
    elseif suffix == "HEAL" then
        local amount, critical = select(paramStart, ...)  
        
        --don't record crit heals but set the critable-flag
        if critical then 
            spell.critical = true
        else
            local min = spell.amountMin
            if not min or min > amount then
               spell.amountMin = amount
            end
            
            local max = spell.amountMax
            if not max or max < amount then
               spell.amountMax = amount
            end
        end
        
    --information concerning spell misses (avoidance flags)
    elseif suffix == "MISSED" then
        local missType = select(paramStart, ...)  
        
        if missType == "DODGE" then
            spell.dodge = true
        elseif missType == "PARRY" then
            spell.parry = true
        elseif missType == "REFLECT" then
            spell.reflect = true
        elseif missType == "RESIST" then
            spell.resist = true
        elseif missType ~= "MISS" then
            spell.missing = true
        end
        
    --information on generel spells which have an amount parameter (leech, drain and energize)
    elseif amount_suffixes[suffix] then
        local amount, critical = select(paramStart, ...)  
        
        local min = spell.amountMin
        if not min or min > amount then
           spell.amountMin = amount
        end
        
        local max = spell.amountMax
        if not max or max < amount then
           spell.amountMax = amount
        end
    end
        
    mobs[zone][NPCID].spells[spellID] = spell
end

local function showSpellTooltip()
	local this = _G.this
	GameTooltip:ClearLines();
	GameTooltip:SetOwner(this, "ANCHOR_RIGHT");
    if this.spellID ~= 0 then
        GameTooltip:SetHyperlink(("spell:%s"):format(this.spellID))
    else
        GameTooltip:AddLine(("|cffffffff%s|r"):format("Melee Attack"))
    end	
	GameTooltip:AddLine(" ")
	if this.spellID ~= 0 then GameTooltip:AddLine(("|cffffffffSpell ID #|r|cffffff00%s|r"):format(this.spellID)) end
    
    local spell = this.spell
    if spell.amountMin then
		GameTooltip:AddLine(("|cffffff00%s-%s %s|r"):format(spell.amountMin, spell.amountMax, spell.schoolName or ""))
	elseif spell.schoolName then
		GameTooltip:AddLine(("|cffffffffSchool|r |cffffff00%s|r"):format(spell.schoolName))
	end
    if spell.minSpeed then GameTooltip:AddLine(("|cffffffffEvery|r |cffffff00%.1f-%.1f seconds|r"):format(spell.minSpeed, spell.maxSpeed)) end
    
    if spell.missing then GameTooltip:AddLine(("|cffffffffCan miss|r")) end
    if spell.resist then GameTooltip:AddLine(("|cffffffffCan full resist|r")) end
    if spell.reflect then GameTooltip:AddLine(("|cffffffffCan be reflected|r")) end
    if spell.dodge then GameTooltip:AddLine(("|cffffffffCan be dodged|r")) end
    if spell.parry then GameTooltip:AddLine(("|cffffffffCan be parried|r")) end
    
    if spell.resisted then GameTooltip:AddLine(("|cffffffffCan be resisted|r")) end
    if spell.blocked then GameTooltip:AddLine(("|cffffffffCan be blocked|r")) end
    
    if spell.dispellable then GameTooltip:AddLine(("|cffffffffCan be dispelled|r")) end
    if spell.stealable then GameTooltip:AddLine(("|cffffffffCan be stolen|r")) end
    if spell.interruptable then GameTooltip:AddLine(("|cffffffffCan be interrupted|r")) end
    
    if spell.critical then GameTooltip:AddLine(("|cffffffffCan crit|r")) end
    if spell.glancing then GameTooltip:AddLine(("|cffffffffCan crush|r")) end
    if spell.crushing then GameTooltip:AddLine(("|cffffffffCan glance|r")) end
    
	--[[
    if spell.events then
        for event, ev in pairs(spell.events) do	
            GameTooltip:AddLine(("%s"):format(event))
        end
    end
	]]--
    
	GameTooltip:Show()
	this.highlight:Show()
end

local function hideTooltip()
	GameTooltip:Hide()
	this.highlight:Hide()
end

local function addLink()
	if IsShiftKeyDown() and ChatFrameEditBox:IsShown() then
		ChatFrameEditBox:Insert(GetSpellLink(this.spellID));
	end
end

local function SelectSpell(widget, event, value)
	widget:ReleaseChildren()

	local zone, mobID = ("\001"):split(value)
	
	if not mobID then return end
	local zoneData = MobSpells.db.global.mobs[zone]
	local data = zoneData[tonumber(mobID)]
	if not data then return end
	
	local sf = AceGUI:Create("ScrollFrame")
	sf:SetLayout("Flow")
	
	local heading1 = AceGUI:Create("Heading")
	heading1:SetText(zone.. ": " .. data.name)
	heading1.width = "fill"
	sf:AddChild(heading1)
	
	for v, spell in pairs(data.spells) do	
		if type(spell) == "number" then
			data.spells[v] = {school = spell}
			spell = data.spells[v]
		end
        local school = spell.school
        local name, rank, icon
        if v == 0 then
            name = "Melee Attack"
            rank = nil
            icon = [[Interface\Icons\INV_SWORD_06]]
        else
            name, rank, icon = GetSpellInfo(v)
        end
        local sFrame = AceGUI:Create("SimpleGroup")
        
		local label = AceGUI:Create("Label")
		if rank and #rank > 0 then
			label:SetText((" [%s (%s)]"):format(name, rank))
		else
			label:SetText((" [%s]"):format(name))
		end
		local colors = COMBATLOG_DEFAULT_COLORS.schoolColoring[school] or COMBATLOG_DEFAULT_COLORS.schoolColoring.default
        if colors then
            label.label:SetTextColor(colors.r, colors.g, colors.b, colors.a)
        end
		label:SetWidth(300)
		local p, h, f = label.label:GetFont()
		label.label:SetFont("Fonts\\ARIALN.TTF", 15, f)
		label.image:SetHeight(22)
		label.image:SetWidth(22)
		label.frame:EnableMouse(true)
        
		label.frame.spellID = v
        label.frame.spell = spell        
        
		label.frame:SetScript("OnEnter", showSpellTooltip)
		label.frame:SetScript("OnLeave", hideTooltip)
		label.frame:SetScript("OnMouseDown", addLink)
		label.frame.highlight = label.frame:CreateTexture("Highlight", "OVERLAY")
		label.frame.highlight:SetTexture("Interface\\Buttons\\UI-PlusButton-Hilight")
		label.frame.highlight:SetAllPoints()
		label.frame.highlight:SetBlendMode("ADD")
		label.frame.highlight:Hide()

        if icon then
            label:SetImage(icon)
        end
        
        sFrame:AddChild(label)
		sf:AddChild(label)
		
		--[[
        --placeholder button for methods like delete from database and post spellinfo to chat
		-- commented out since it's to-do
        local btn = AceGUI:Create("Button")
        btn:SetText("G")
        btn:SetHeight(22)
        btn:SetWidth(22)
		]]--
        
        -- sf:AddChild(btn)
	end
	
	widget:AddChild(sf)
end

function MobSpells:CreateFrame()
	if frame then return end
	
	local f = AceGUI:Create("Frame")
	f:SetTitle("Mob Spells")
	f:SetStatusText("")
	f:SetLayout("Flow")
	f:SetWidth(650)
	f:SetHeight(450)
	f.width = "fill"
	
	local dropdown = AceGUI:Create("Dropdown")
	dropdown:SetText("NPCs")
	dropdown:SetLabel("Search")
	dropdown.list = {NPC = "NPCs", SPELL = "Spells"}
	dropdown:SetWidth(100)
	dropdown:SetCallback("OnValueChanged",function(widget,event,value) MobSpells.searchIndex = value end )
	f:AddChild(dropdown)

	local edit = AceGUI:Create("EditBox")
	edit:SetText("")
	edit:SetWidth(200)
	edit:SetLabel("Keywords")
	edit:SetCallback("OnEnterPressed",function(widget,event,text) MobSpells:Filter(text) end )
	edit:SetCallback("OnTextChanged",function(widget,event,text) MobSpells:Filter(text) end )
	f.editBox = edit
	f:AddChild(edit)
	
	local t = AceGUI:Create("TreeGroup")
	t:SetLayout("Fill")
	t:SetCallback("OnGroupSelected", SelectSpell )
	t.width = "fill"
	t.height = "fill"
	
	f:AddChild(t)
	f.tree = t

	frame = f
end

local function clearTree(t)
	for k, v in pairs(t) do
		if type(v) == "table" then
			t[k] = clearTree(v)
		else
			t[k] = nil
		end	
	end
	return del(t)
end

local function comp(a, b)
	if a[1] > b[1] then
		return false
	elseif a[1] < b[1] then
		return true
	else
		if a[2] > b[2] then
			return false
		elseif a[2] < b[2] then
			return true
		else
			return false
		end
	end
end

local t = new()
function MobSpells:BuildTree()
	t = clearTree(t)
	t = new()
	local zones = new()
	for zone, mobList in pairs(mobs) do
		tinsert(zones, zone)
	end
	table_sort(zones)
	for i = 1, #zones do
		local zone = zones[i]
		local mobList = mobs[zone]
		local tzone = newHash()
		tzone.text = zone
		tzone.value = zone
		local children = new()
		tzone.children = children
		local mobNames = new()
		for mobID, data in pairs(mobList) do
			tinsert(mobNames, new(data.name, mobID))
		end
		table_sort(mobNames, comp)
		for i = 1, #mobNames do
			local t = mobNames[i]
			local mobID = t[2]
			local data = mobList[mobID]
			local mob = new()
			mob.text = data.name
			mob.value = mobID
			tinsert(children, mob)
		end
		for i = 1, #mobNames do
			mobNames[i] = del(mobNames[i])
		end
		mobNames = del(mobNames)
		tinsert(t, tzone)
	end	
	zones = del(zones)
	return t
end

function MobSpells:Filter(text)
	local t = self:BuildTree()
	-- Iterate zones
	local zOffset = 0
	for i = 1, #t do
		local v = t[i-zOffset]
		local offset = 0
		for j = 1, #v.children do
			local child = v.children[j-offset]
			local useMob = false
			if #text == 0 then
				useMob = true
			end
			if self.searchIndex == "NPC" and child.text:lower():match(text:lower()) then
				useMob = true
			elseif self.searchIndex == "SPELL" then
				for k, v in pairs(self.db.global.mobs[v.value][child.value].spells) do
					local name = GetSpellInfo(k)
					if name:lower():match(text:lower()) then
						useMob = true
						break
					end
				end
			end
			if not useMob then
				tremove(v.children, j - offset)
				offset = offset + 1
			end
		end
		if #v.children == 0 then
			tremove(t, i - zOffset)
			zOffset = zOffset + 1
		end
	end
	frame.tree:SetTree(t)
end

function MobSpells:GetSpells(unit)
	local name = UnitName(unit)
	MobSpells.searchIndex = "NPC"
	self:Show()
	frame.editBox:SetText(name)
	self:Filter(name)
	local guid = UnitGUID(unit)
	local NPCID = tonumber(guid:sub(-12,-7),16)
	local zone = GetZoneText()
    if IsInInstance() and GetInstanceDifficulty() == 2 then 
        zone = zone.." [Heroic]" 
    end
	
	frame.tree:SelectByValue(zone .. "\001" .. NPCID)
	-- frame.tree:SetSelected
end

function MobSpells.GetTargetSpells()
	MobSpells:GetSpells(UnitExists("target") and "target" or "mouseover")
end

function MobSpells:ReportTargetSpells()
	local unit = UnitExists("target") and "target" or "mouseover"
	local guid = UnitGUID(unit)
	local NPCID = tonumber(guid:sub(-12,-7),16)
	local name = UnitName(unit)
	local zone = GetZoneText()
    if IsInInstance() and GetInstanceDifficulty() == 2 then 
        zone = zone.." [Heroic]" 
    end
	
	local s = name .. " casts: "
	local added = false
	if mobs and mobs[zone] and mobs[zone][NPCID] and mobs[zone][NPCID].spells then
		for k, v in pairs(mobs[zone][NPCID].spells) do
			local link = GetSpellLink(k)
			if link and #link > 0 then
				if #s + #link > 255 then
					SendChatMessage(s, "RAID")
					s = ""
				end
				s = s .. link
				added = true
			end
		end
		if #s > 0 and added then
			SendChatMessage(s, GetNumRaidMembers() > 0 and "RAID" or "PARTY")
		elseif not added then
			ChatFrame1:AddMessage("No recorded spells for " .. name)
		end
	end
end
function MobSpells:Show(input)
	if input == "tooltip" then
		self.db.profile.showTooltips = not self.db.profile.showTooltips
		if self.db.profile.showTooltips then
			self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
			self:Print("Now showing abilities in tooltip")
		else
			self:UnregisterEvent("UPDATE_MOUSEOVER_UNIT")
			self:Print("NOT showing abilities in tooltip")
		end
		return
	elseif input == "report" then
		self:ReportTargetSpells()
		return
	end

	self:CreateFrame()
	frame.tree:SetTree(self:BuildTree())
	frame:Show()
end
